package rendezvous;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;

/**
 * Class for directing rendez-vous
 * @author Miroslav Hendrych
 */
public class RendezVous {
    private HashMap <Long, LinkedList<EntryCall>> calls;
    private HashMap <Long, HashMap<String, IEntry>> entry;
    private static IAcceptCondition trueCondition = new AcceptConditionTrue();

    public RendezVous() {
        calls = new HashMap<Long, LinkedList<EntryCall>>();
        entry = new HashMap<Long, HashMap<String, IEntry>>();
    }

    /**
     * Obyčejný akcept na všech entry
     */
    public void makeAccept() {
        Long id = Thread.currentThread().getId();
        LinkedList<EntryCall> list = calls.get(id);
        EntryCall call = null;
        synchronized (list) {
            while (list.isEmpty()) {
                try {
                    list.wait();
                } catch (InterruptedException ex) {
                    ex.printStackTrace();
                }
            }
            call = list.pollFirst();
        }
        HashMap<String, IEntry> serverEntry = entry.get(id);
        IEntry calledEntry = serverEntry.get(call.getEntryName());
        Object result = calledEntry.entry(call.getParams());
        call.setResult(result);
    }

    /**
     * Nepodmíněný akcept na omezené skupině entry
     * @param acceptedEntry Názvy entry
     */
    public void makeAccept(String[] acceptedEntry) {
        /* Převede zadanou skupinu na podmíněné akceptování, kde podmínky jsou vždy splněné */
        HashMap<String, IAcceptCondition> acceptedEntryHM = new HashMap<String, IAcceptCondition>();
        for (String accEntry : acceptedEntry) {
            acceptedEntryHM.put(accEntry, trueCondition);
        }
        makeAccept(acceptedEntryHM, 0);
    }

    /**
     * Podmíněný akcept na omezené skupině entry
     * @param acceptedEntry Názvy entry a jejich podmínky
     */
    public void makeAccept(HashMap<String, IAcceptCondition> acceptedEntry) {
        makeAccept(acceptedEntry, 0);
    }

    /**
     * Podmíněný akcept na omezené skupině entry časově omezený
     * @param acceptedEntry Názvy entry a jejich podmínky
     * @param timeOut čas v milisekundách větší než 0, po kterém je pokus o akcept zrušen (0=neomezeně)
     * @return Vrací true, pokud došlo k akceptu nějakého entry
     */
    public boolean makeAccept(HashMap<String, IAcceptCondition> acceptedEntry, long timeOut) {
        /* systémový čas, kdy budeme končit */
        long konec = System.currentTimeMillis() + timeOut;

        Long id = Thread.currentThread().getId();
        LinkedList<EntryCall> list = calls.get(id);
        EntryCall call = null;
        int index = 0;

        synchronized (list) {
            do {
                /* projdeme celou frontu od poslední prohledávané pozice
                 * tím se vyhneme prozkoumávání již vyhodnocených volání */
                for (Iterator<EntryCall> it = list.listIterator(index); it.hasNext();) {
                    EntryCall ecall = it.next();
                    /* otestování, zda entry chceme vůbec akceptovat
                       a pokud ano, tak zda je splněná podmínka */
                    if ( acceptedEntry.containsKey(ecall.getEntryName()) &&
                         acceptedEntry.get(ecall.getEntryName()).isAcceptable()) {
                        call = ecall;
                        list.remove(index);
                        break;
                    }
                    index++;
                }

                /* prošli jsme celou frontu a neakceptovali jsme žádné call
                 * seznam mohl být prázdný nebo neobsahoval žádné akceptovatelné call
                 */
                if (call == null) {
                    try {
                        /* jedná se o časově omezený pokus o akcept */
                        if (timeOut>0) {
                            /* vypočítáme zbývající dobu, musí být kladná */
                            long left = Math.max(0l, konec - System.currentTimeMillis());
                            if (left>0)
                                list.wait(left);  // ještě máme čas, počkáme
                            else
                                return false;  // čas vypršel, vracíme false
                        } else  // časově neomezené volání, čekáme na příchozí call
                            list.wait();
                    } catch (InterruptedException ex) {
                        ex.printStackTrace();
                    }
                }
            } while (call == null);
        }

        /* vyhledáme příslušné tělo acceptu a zavoláme ho */
        HashMap<String, IEntry> serverEntry = entry.get(id);
        IEntry calledEntry = serverEntry.get(call.getEntryName());
        Object result = calledEntry.entry(call.getParams());
        
        /* uložíme návratovou hodnotu, metoda zároveň probudí klientské vlákno */
        call.setResult(result);
        return true;
    }

    /**
     * 
     * @param thread ID serverového vlákna
     * @param entry název volaného entry
     * @param params parametry
     * @return
     */
    public Object call(long thread, String entry, Object params) {
        LinkedList<EntryCall> list = calls.get(thread);
        if (list==null)
            return null;
        EntryCall entryCall = new EntryCall(thread, entry, params);
        synchronized (list) {
            list.add(entryCall);
            list.notify();
        }
        return entryCall.getResult();
    }

    /**
     * Zaregistrování serverového vlákna
     * @param id ID vlákna
     * @return vrací true v případě úspěchu
     */
    public synchronized boolean registerServer(Long id) {
        if (!calls.containsKey(id)) {
            calls.put(id, new LinkedList<EntryCall>());
            entry.put(id, new HashMap<String, IEntry>());
            return true;
        }
        return false;
    }

    /**
     * Zaregistrování sebe sama jako serverové vlákno
     * @return vrací true v případě úspěchu
     */
    public synchronized boolean registerServer() {
        Long id = Thread.currentThread().getId();
        return registerServer(id);
    }

    /**
     * Odregistrování serverového vlákna, všech svých entry a zrušení probíhajících call
     */
    public synchronized void unregisterServer() {
        Long id = Thread.currentThread().getId();
        if (calls.containsKey(id)) {
            LinkedList<EntryCall> list = calls.get(id);
            synchronized (list) {
               for (Iterator<EntryCall> it = list.iterator(); it.hasNext(); ) {
                    EntryCall call = it.next();
                    call.setResult(null);
                }
                list.clear();
            }
            entry.remove(id);
            calls.remove(id);
        }
    }

    /**
     * Zaregistrování entry pro serverové vlákno
     * @param id ID serverového vlákna
     * @param name název entry
     * @param entryClass rozhraní pro tělo akceptu
     * @return Vrací true v případě úspěchu
     */
    public synchronized boolean registerEntry(Long id, String name, IEntry entryClass) {
        if (!calls.containsKey(id))
            if (!registerServer())
                return false;
        HashMap<String, IEntry> serverEntry = entry.get(id);
        if (serverEntry.containsKey(name))
            return false;
        serverEntry.put(name, entryClass);
        return true;
    }

    /**
     * Zaregistrování entry pro sebe jako serverové vlákno
     * @param name název entry
     * @param entryClass rozhraní pro tělo akceptu
     * @return Vrací true v případě úspěchu
     */
    public synchronized boolean registerEntry(String name, IEntry entryClass) {
        Long id = Thread.currentThread().getId();
        return registerEntry(id, name, entryClass);
    }

    /**
     * Vrací ID vláken, které mají registrované takto pojmenované entry
     * @param entryName název entry
     * @return Vrací pole id vláken
     */
    public synchronized Long[] getServersByEntryName(String entryName) {
        HashSet<Long> servers = new HashSet<Long>();
        for (Iterator<Entry<Long, HashMap<String, IEntry>>> it = entry.entrySet().iterator(); it.hasNext(); ) {
            Entry<Long, HashMap<String, IEntry>> e = it.next();
            if (e.getValue().containsKey(entryName))
                servers.add(e.getKey());
        }
        return servers.toArray(new Long[0]);
    }

    /**
     * Vrací názvy entry, které má registrované dané vlákno
     * @param id ID serverového vlákna
     * @return Vrací pole názvů entry
     */
    public synchronized String[] getEntryByServerID(Long id) {
        return entry.get(id).keySet().toArray(new String[0]);
    }

    /**
     * Pro jednoduší vytváření podmínek tato metoda vrací instanci podmínky, která je vždy splněná
     * @return Instance objektu pro rozhodovací podmínky akceptování
     */
    public static IAcceptCondition getTrueCondition() {
        return trueCondition;
    }
}
